%{
/*
       Copyright 1989, 1990, 1993, 1994 Digital Equipment Corporation
                        All Rights Reserved

Permission to use, copy, and modify this software and its documentation is
hereby granted only under the following terms and conditions.  Both the
above copyright notice and this permission notice must appear in all copies
of the software, derivative works or modified versions, and any portions
thereof, and both notices must appear in supporting documentation.

Users of this software agree to the terms and conditions set forth herein,
and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
free right and license under any changes, enhancements or extensions 
made to the core functions of the software, including but not limited to 
those affording compatibility with other hardware or software 
environments, but excluding applications which incorporate this software.
Users further agree to use their best efforts to return to Digital any
such changes, enhancements or extensions that they make and inform Digital
of noteworthy uses of this software.  Correspondence should be provided
to Digital at:

                      Director of Licensing
                      Western Research Laboratory
                      Digital Equipment Corporation
                      250 University Avenue
                      Palo Alto, California  94301

This software may be distributed (but not offered for sale or transferred
for compensation) to third parties, provided such third parties agree to
abide by the terms and conditions of this notice.

THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
 */

#if !defined(LINT) && !defined(CODECENTER)
static char rcsid[] = "$Id: screentab.y,v 2.6 1995/08/19 09:55:23 vixie Exp $";
#endif

/*
 * screentab.y
 *
 * yacc grammar for screentab
 *
 * Modification history:
 *
 * 19 December 1988	Jeffrey Mogul/DECWRL
 *	Created.
 * Modified extensively since then; see RCS.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <net/gw_screen.h>

#include <stdio.h>
#include <stdlib.h>

#include "screentab.h"
#include "funcproto.h"

extern int yylineno;
extern char linebuf[];

extern int semantic_errors;

extern char *sourcefile;
int	defaction = ASACTION_REJECT;

struct NetmaskData curnetmask;
struct ActionSpec  curaction;
struct ActionSpec  curactionreverse;
struct AddrSpecX *asxp;
struct PortSpecX *psxp;
struct PortValue pval;
struct in_addr_byte ia;
int hexnum;
int masklen;
int osp_flags;

static struct in_addr MaskLenToVal();

/*
 * flags-not values are, during the parse, encoded in the
 * high-order byte of a 16-bit value.
 */
#define	FLAGSNOTSHIFT	(8)
#define	FLAGMASK	(0xFF)
#define	YESFLAGS(flgs)	(flgs & FLAGMASK)
#define	NOFLAGS(flgs)	((flgs >> FLAGSNOTSHIFT) & FLAGMASK)

struct ObjectSpec *MakeOSpec();
%}

%start	ScreenTabFile

/*
 * Reserved words
 */
%token	ACCEPT ADDR ADDRNOT AND ANY BETWEEN DEFAULT FOR FROM
%token	HOST HOSTNOT ICMP INFOTYPE IS
%token	LOG MASK MASKLEN MULTICAST NET NETNOT NETMASK NOTIFY
%token	PORT PORTNOT PROTO PROTONOT
%token	RANGE REJECTIT RESERVED SUBNET SUBNETNOT
%token	THROUGH TCP TO TYPE TYPENOT UDP XSERVER

/*
 * Reserved words for TCP header flags
 */
%token	FLAGS FLAGSNOT ACK FIN PUSH RST SYN URG
%token	INITIALF CONTINUING

/*
 * string-valued tokens
 */
%token	NAME	LITERAL

/*
 * numeric-valued tokens (but they are passed as strings)
 */
%token	DECNUMBER HEXNUMBER

/*
 * punctuation
 */
%token	DOT
%token	SEMICOLON

%%	/* rules section */

ScreenTabFile	:	RuleList
		;

RuleList	:	SubnetSpec
		|	DefaultSpec
		|	PairSpec
		|	RuleList SubnetSpec
		|	RuleList PairSpec
		|	RuleList DefaultSpec
		;

SubnetSpec	:	FOR Netval NETMASK IS Maskval SEMICOLON
			{
			    curnetmask.network.s_addr = $2;
			    curnetmask.mask.s_addr = $5;
			    StoreNetmask(&curnetmask);
			}
		;

DefaultSpec	:	DEFAULT Decision SEMICOLON
			{
			    defaction = $2;
			}
		;

PairSpec	:	FROM Object TO Object Decision SEMICOLON
			{
			    struct ObjectSpec *from = (struct ObjectSpec *)$2,
					      *to = (struct ObjectSpec *)$4;
			    curaction.from = *from;
			    free(from);
			    curaction.to = *to;
			    free(to);
			    curaction.action = $5;
			    StoreAction(&curaction);
			}
		|	FROM Object TO Object FlagSpec Decision SEMICOLON
			{
			    struct ObjectSpec *from = (struct ObjectSpec *)$2,
					      *to = (struct ObjectSpec *)$4;
			    curaction.from = *from;
			    curaction.from.tcpyesflags = YESFLAGS($5);
			    curaction.from.tcpnoflags = NOFLAGS($5);
			    free(from);
			    curaction.to = *to;
			    curaction.to.tcpyesflags = YESFLAGS($5);
			    curaction.to.tcpnoflags = NOFLAGS($5);
			    free(to);
			    curaction.action = $6;
			    StoreAction(&curaction);
			}
		|	BETWEEN Object AND Object Decision SEMICOLON
			{
			    struct ObjectSpec *from = (struct ObjectSpec *)$2,
					      *to = (struct ObjectSpec *)$4;
			    curaction.from = *from;
			    free(from);
			    curaction.to = *to;
			    free(to);
			    curaction.action = $5;
			    StoreAction(&curaction);
			    curactionreverse.from = curaction.to;
			    curactionreverse.to = curaction.from;
			    curactionreverse.action = curaction.action;
			    StoreAction(&curactionreverse);
			}
		|	BETWEEN Object AND Object FlagSpec Decision SEMICOLON
			{
			    struct ObjectSpec *from = (struct ObjectSpec *)$2,
					      *to = (struct ObjectSpec *)$4;
			    curaction.from = *from;
			    curaction.from.tcpyesflags = YESFLAGS($5);
			    curaction.from.tcpnoflags = NOFLAGS($5);
			    free(from);
			    curaction.to = *to;
			    curaction.to.tcpyesflags = YESFLAGS($5);
			    curaction.to.tcpnoflags = NOFLAGS($5);
			    free(to);
			    curaction.action = $6;
			    StoreAction(&curaction);
			    curactionreverse.from = curaction.to;
			    curactionreverse.to = curaction.from;
			    curactionreverse.action = curaction.action;
			    StoreAction(&curactionreverse);
			}
		;

Decision	:	ACCEPT
			{ $$ = ASACTION_ACCEPT; }
		|	ACCEPT LOG
			{ $$ = ASACTION_ACCEPT|ASACTION_LOG; }
		|	REJECTIT
			{ $$ = ASACTION_REJECT; }
		|	REJECTIT LOG
			{ $$ = ASACTION_REJECT|ASACTION_LOG; }
		|	REJECTIT NOTIFY
			{ $$ = ASACTION_REJECT|ASACTION_NOTIFY; }
		|	REJECTIT NOTIFY LOG
			{ $$ = ASACTION_REJECT|ASACTION_NOTIFY|ASACTION_LOG; }
		|	REJECTIT LOG NOTIFY
			{ $$ = ASACTION_REJECT|ASACTION_NOTIFY|ASACTION_LOG; }
		;

Object		:	ObjAddress
			{
			    $$ = (YYSTYPE)MakeOSpec(
						(struct AddrSpecX *)$1,
						NULL);
			}
		|	ObjPort
			{
			    $$ = (YYSTYPE)MakeOSpec(
						NULL,
						(struct PortSpecX *)$1);
			}
		|	ObjAddress ObjPort
			{
			    $$ = (YYSTYPE)MakeOSpec(
						(struct AddrSpecX *)$1,
						(struct PortSpecX *)$2);
			}
		;

AddrOrNot	:	ADDRNOT
			{ $$ = OSF_NOTADDR; }
		|	ADDR
			{ $$ = OSF_DEFAULT; }
		;

HostOrNot	:	HOSTNOT
			{ $$ = OSF_NOTADDR; }
		|	HOST
			{ $$ = OSF_DEFAULT; }
		;

NetOrNot	:	NETNOT
			{ $$ = OSF_NOTADDR; }
		|	NET
			{ $$ = OSF_DEFAULT; }
		;

SubnetOrNot	:	SUBNETNOT
			{ $$ = OSF_NOTADDR; }
		|	SUBNET
			{ $$ = OSF_DEFAULT; }
		;

PortOrNot	:	PORTNOT
			{ $$ = OSF_NOTPORT; }
		|	PORT
			{ $$ = OSF_DEFAULT; }
		;

TypeOrNot	:	TYPENOT
			{ $$ = OSF_NOTPORT; }
		|	TYPE
			{ $$ = OSF_DEFAULT; }
		;

ProtoOrNot	:	PROTONOT
			{ $$ = OSF_NOTPROTO; }
		|	PROTO
			{ $$ = OSF_DEFAULT; }
		;

ObjAddress	:	NetOrNot Netval
			{
			    asxp = (struct AddrSpecX *)malloc(sizeof(*asxp));
			    
			    if ($2 == INADDR_ANY) {
				asxp->aspec.addrtype = ASAT_ANY;
				asxp->aspec.aval.host.s_addr = 0;
			    }
			    else {
				asxp->aspec.addrtype = ASAT_NET;
				VerifyNetVal($2);
				asxp->aspec.aval.network.s_addr = $2;
			    }
			    asxp->flags = $1;
			    $$ = (YYSTYPE)asxp;
			}
		|	SubnetOrNot Subnetval
			{
			    asxp = (struct AddrSpecX *)malloc(sizeof(*asxp));
			    
			    if ($2 == INADDR_ANY) {
				asxp->aspec.addrtype = ASAT_ANY;
				asxp->aspec.aval.host.s_addr = 0;
			    }
			    else {
				asxp->aspec.addrtype = ASAT_SUBNET;
				asxp->aspec.aval.subnet.s_addr = $2;
			    }
			    asxp->flags = $1;
			    $$ = (YYSTYPE)asxp;
			}
		|	HostOrNot Hostval
			{
			    asxp = (struct AddrSpecX *)malloc(sizeof(*asxp));
			    
			    if ($2 == INADDR_ANY) {
				asxp->aspec.addrtype = ASAT_ANY;
				asxp->aspec.aval.host.s_addr = 0;
			    }
			    else {
				asxp->aspec.addrtype = ASAT_HOST;
				asxp->aspec.aval.host.s_addr = $2;
			    }
			    asxp->flags = $1;
			    $$ = (YYSTYPE)asxp;
			}
		|	AddrOrNot Hostval MASK Maskval
			{
			    asxp = (struct AddrSpecX *)malloc(sizeof(*asxp));
			    
			    asxp->aspec.addrtype = ASAT_ADDR;
			    asxp->aspec.aval.addr.s_addr = $2;
			    asxp->aspec.mask.s_addr = $4;
			    if (asxp->aspec.aval.addr.s_addr &
				~asxp->aspec.mask.s_addr) {
				yyerror("host part not zero");
				semantic_errors++;
			    }
			    asxp->flags = $1;
			    $$ = (YYSTYPE)asxp;
			}
		|	AddrOrNot Hostval MASKLEN Masklen
			{
			    asxp = (struct AddrSpecX *)malloc(sizeof(*asxp));
			    
			    asxp->aspec.addrtype = ASAT_ADDR;
			    asxp->aspec.aval.addr.s_addr = $2;
			    asxp->aspec.mask = MaskLenToVal($4);
			    if (asxp->aspec.aval.addr.s_addr &
				~asxp->aspec.mask.s_addr) {
				yyerror("host part not zero");
				semantic_errors++;
			    }
			    asxp->flags = $1;
			    $$ = (YYSTYPE)asxp;
			}
		|	ANY
			{
			    asxp = (struct AddrSpecX *)malloc(sizeof(*asxp));
			    
			    asxp->aspec.addrtype = ASAT_ANY;
			    asxp->aspec.aval.host.s_addr = 0;
			    asxp->flags = OSF_DEFAULT;
			    $$ = (YYSTYPE)asxp;
			}
		;

ObjPort		:	ICMP TypeOrNot Icmptype
			{
			    psxp = (struct PortSpecX *)malloc(sizeof(*psxp));

			    psxp->pspec.proto = IPPROTO_ICMP;
			    psxp->pspec.pval.code = $3;
			    psxp->flags = $2;
			    $$ = (YYSTYPE)psxp;
			}
		|	TCP PortOrNot Portval
			{
			    psxp = (struct PortSpecX *)malloc(sizeof(*psxp));

			    psxp->pspec.proto = IPPROTO_TCP;
			    psxp->pspec.pval.ports =
						*((struct PortValue *)($3));
			    psxp->flags = $2;
			    $$ = (YYSTYPE)psxp;
			}
		|	UDP PortOrNot Portval
			{
			    psxp = (struct PortSpecX *)malloc(sizeof(*psxp));

			    psxp->pspec.proto = IPPROTO_UDP;
			    psxp->pspec.pval.ports =
			    			*((struct PortValue *)($3));
			    psxp->flags = $2;
			    $$ = (YYSTYPE)psxp;
			}
		|	ProtoOrNot Protoval
			{
			    psxp = (struct PortSpecX *)malloc(sizeof(*psxp));

			    switch (psxp->pspec.proto = $2) {
			    case IPPROTO_ICMP:
				psxp->pspec.pval.code = ICMPV_ANY;
				break;
			    case IPPROTO_TCP: /*FALLTHROUGH*/
			    case IPPROTO_UDP:
				psxp->pspec.pval.ports.first = IPPORT_FIRST;
				psxp->pspec.pval.ports.last = IPPORT_LAST;
				break;
			    default:
				break;
			    }
			    psxp->flags = $1;
			    $$ = (YYSTYPE)psxp;
			}
		;

/* Every rule after here returns an integer value */

Netval		:	NAME
			{
			    $$ = NetNameToNumber((char*)$1);
			 }
		|	ANY
			{ $$ = INADDR_ANY; }
		|	DottedQuad
			{ $$ = $1; }
		;

Maskval		:	NAME
			{ $$ = MaskNameToNumber((char*)$1); }
		|	DottedQuad
			{ $$ = $1; }
		;

Subnetval	:	NAME
			{ $$ = SubnetNameToNumber((char*)$1); }
		|	ANY
			{ $$ = INADDR_ANY; }
		|	DottedQuad
			{ $$ = $1; }
		;

Hostval		:	NAME
			{ $$ = HostNameToNumber((char*)$1); }
		|	ANY
			{ $$ = INADDR_ANY; }
		|	DottedQuad
			{ $$ = $1; }
		;

DottedQuad	:	Number DOT Number DOT Number DOT Number
			{
			    ia.addr_un.S_un_b.s_b1 = $1;
			    ia.addr_un.S_un_b.s_b2 = $3;
			    ia.addr_un.S_un_b.s_b3 = $5;
			    ia.addr_un.S_un_b.s_b4 = $7;
			    $$ = (u_int32)ia.s_addr_i;	/* network order */
			}
		;

Icmptype	:	NAME
			{ $$ = ICMPNameToNumber((char*)$1); }
		|	ANY
			{ $$ = ICMPV_ANY; }
		|	INFOTYPE
			{ $$ = ICMPV_INFOTYPE; }
		|	Number
			{ $$ = $1; }
		;

Portval		:	NAME
			{
			    pval.first = PortNameToNumber((char*)$1);
			    pval.last = pval.first;
			    $$ = (YYSTYPE)(&pval);
			}
		|	ANY
			{
			    pval.first = IPPORT_FIRST;
			    pval.last = IPPORT_LAST;
			    $$ = (YYSTYPE)(&pval);
			}
		|	RESERVED
			{
			    pval.first = IPPORT_FIRST;
			    pval.last = IPPORT_RESERVED - 1;
			    $$ = (YYSTYPE)(&pval);
			}
		|	XSERVER
			{
			    pval.first = XSERVERPORT_MIN;
			    pval.last = XSERVERPORT_MAX;
			    $$ = (YYSTYPE)(&pval);
			}
		|	RANGE Number THROUGH Number
			{
			    pval.first = $2;
			    pval.last = $4;
			    $$ = (YYSTYPE)(&pval);
			}
		|	Number
			{
			    pval.first = $1;
			    pval.last = pval.first;
			    $$ = (YYSTYPE)(&pval);
			}
		;

Protoval	:	NAME
			{ $$ = ProtoNameToNumber((char*)$1); }
		|	Number
			{ $$ = $1; }
		|	TCP
			{ $$ = IPPROTO_TCP; }
		|	UDP
			{ $$ = IPPROTO_UDP; }
		|	ICMP
			{ $$ = IPPROTO_ICMP; }
		;


FlagSpec	:	FLAGS FlagVal
			{ $$ = $2; }
		|	FLAGSNOT FlagVal
			{ $$ = ($2) << FLAGSNOTSHIFT; }
		|	INITIALF
			{ $$ = TH_SYN | (TH_ACK << FLAGSNOTSHIFT); }
		|	CONTINUING
			{ $$ = TH_ACK; }
		|	FlagSpec FlagSpec
			{ $$ = $1 | $2; }
		;

FlagVal		:	Number
		|	Flags
		;

Flags		:	Flags Flag
			{ $$ = $1 | $2; }
		|	Flag
		;

Flag		:	ACK
			{ $$ = TH_ACK; }
		|	FIN
			{ $$ = TH_FIN; }
		|	PUSH
			{ $$ = TH_PUSH; }
		|	RST
			{ $$ = TH_RST; }
		|	SYN
			{ $$ = TH_SYN; }
		|	URG
			{ $$ = TH_URG; }
		;

/* Numeric conversions */
Number		:	HEXNUMBER
			{
			    sscanf((char*)$1, "%x", &hexnum);
			    $$ = hexnum;
			}
		|	DECNUMBER
			{
			    $$ = atoi((char*)$1);
			}
		;

Masklen		:	DECNUMBER
			{
			    masklen = atoi((char*)$1);
			    if (masklen < 0 || masklen > 32) {	/* XXX */
				yyerror("mask length out of range (1..32)");
				semantic_errors++;
				masklen = 0;
			    }
			    $$ = masklen;
			}
		;

%%	/* subroutines */

#if !(defined(ultrix) || defined(__osf__))
int yylineno = 0;
#endif

void
yyerror(s)
char *s;
{
	fflush(stdout);
	fprintf(stderr, "\"%s\", line %d: %s\n", sourcefile, yylineno, s);
	fprintf(stderr, "erroneous line: %s\n", linebuf);
}

void
yywarn(s)
char *s;
{
	fflush(stdout);
	fprintf(stderr, "\"%s\", line %d: %s\n", sourcefile, yylineno, s);
	fprintf(stderr, "questionable line: %s\n", linebuf);
}

static struct in_addr
MaskLenToVal(len)
int len;
{
	struct in_addr val;

	if (len == 32)
		val.s_addr = 0xffffffff;
	else
		val.s_addr = ((1 << len) - 1) << (32 - len);
	return (val);
}

struct ObjectSpec *
MakeOSpec(asxp, psxp)
struct AddrSpecX *asxp;	/* WE FREE THIS if not NULL */
struct PortSpecX *psxp;	/* WE FREE THIS if not NULL */
{
	struct ObjectSpec *osp = (struct ObjectSpec *)malloc(sizeof(*osp));
	
	osp->flags = OSF_DEFAULT;

	if (asxp) {
	    osp->aspec = asxp->aspec;
	    osp->flags |= asxp->flags;
	    free(asxp);
	}
	else {
	    osp->aspec.addrtype = ASAT_ANY;
	    osp->aspec.aval.host.s_addr = 0;
	}

	if (psxp) {
	    osp->pspec = psxp->pspec;
	    osp->flags |= psxp->flags;
	    free(psxp);
	}
	else {
	    osp->pspec.proto = 0;
	    osp->pspec.pval.ports.first = IPPORT_FIRST;
	    osp->pspec.pval.ports.last = IPPORT_LAST;
	}

	osp->tcpyesflags = 0;
	osp->tcpnoflags = 0;

	return(osp);
}

void
VerifyNetVal(nv)
register u_int32 nv;		/* in network order */
{
	register u_int32 i = ntohl(nv);
	register u_int32 netnum;
	
	if (IN_CLASSA(i))
	    netnum = htonl(i&IN_CLASSA_NET);
	else if (IN_CLASSB(i))
	    netnum = htonl(i&IN_CLASSB_NET);
	else if (IN_CLASSC(i))
	    netnum = htonl(i&IN_CLASSC_NET);
	else
	    netnum = i;	/* XXX wrong for Multicast? XXX */

	if (nv != netnum) {
	    yyerror("malformed network number (might be a subnet?)");
	    semantic_errors++;
	}
}
